#![cfg_attr(feature = "nightly", feature(error_in_core))]
#![no_std]

pub mod error {
    use core::fmt::Display;

    use crate::Version;

    #[derive(Debug, PartialEq, Eq)]
    pub enum BaseBlockVerifyError {
        BadSignature,
        IncompatibleVersion(Version),
        IncompatibleMajorVersion,
        IncompatibleMinorVersion,
        BadFileType,
        BadFormat,
        Cluster,
    }

    impl Display for BaseBlockVerifyError {
        fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
            match self {
                BaseBlockVerifyError::BadSignature => f.write_str("Bad Signature"),
                BaseBlockVerifyError::IncompatibleVersion(version) => {
                    f.write_fmt(format_args!("{:?}", version))
                }
                BaseBlockVerifyError::IncompatibleMajorVersion => {
                    f.write_str("Incompatible Major Version")
                }
                BaseBlockVerifyError::IncompatibleMinorVersion => {
                    f.write_str("Incompatible Minor Version")
                }
                BaseBlockVerifyError::BadFileType => f.write_str("Bad File Type (self.kind)"),
                BaseBlockVerifyError::BadFormat => f.write_str("Bad Format"),
                BaseBlockVerifyError::Cluster => f.write_str("Not the correct Cluster"),
            }
        }
    }

    #[cfg(feature = "nightly")]
    impl core::error::Error for BaseBlockVerifyError {}
}

pub mod ptr {

    use bytemuck::{Pod, Zeroable};

    #[repr(transparent)]
    #[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy)]
    pub struct ZstPtr;

    unsafe impl Zeroable for ZstPtr {
        fn zeroed() -> Self {
            Self
        }
    }

    unsafe impl Pod for ZstPtr {}

    impl ZstPtr {
        pub fn ptr(&self) -> *const () {
            core::ptr::addr_of!(*self).cast()
        }

        pub fn ptr_mut(&mut self) -> *mut () {
            core::ptr::addr_of_mut!(*self).cast()
        }
    }

    #[test]
    fn zst_ptr() {
        use core::mem::size_of;

        assert_eq!(size_of::<ZstPtr>(), 0);

        let ptr = ZstPtr;
        let ref r = ptr;

        assert_eq!(ptr.ptr(), r as *const ZstPtr as *const ());
    }
}

pub mod cell_data {
    use core::marker::PhantomData;

    use bitfield::{Bit, BitRange};
    use bitflags::bitflags;
    use bytemuck::{Pod, Zeroable};

    use crate::{ptr::ZstPtr, Offset};

    #[repr(C)]
    #[derive(Debug, Clone, Copy, Zeroable, Pod)]
    pub struct FastEntry {
        offset: Offset,
        name_hint: [u8; 4],
    }

    #[repr(C)]
    #[derive(Debug, Clone, Copy, Zeroable, Pod)]
    pub struct HashEntry {
        offset: Offset,
        name_hash: [u8; 4],
    }

    #[repr(C, packed)]
    #[derive(Debug, Clone, Copy, Zeroable, Pod)]
    pub struct KeyValue {
        name_len: u16,
        data_len: u32,
        data: Offset,
        data_type: u32,
        flags: KeyValueFlags,
        spare: u16,
        value_name: ZstPtr,
    }

    bitflags! {

        #[repr(C)]
        #[derive(Zeroable, Pod)]
        pub struct KeyValueFlags: u16 {
            /// Value name is an ASCII string, possibly an extended ASCII string
            /// (otherwise it is a UTF-16LE string)
            const VALUE_COMP_NAME = 0x1;
            /// Is a tombstone value (the flag is used starting from Insider
            /// Preview builds of Windows 10 "Redstone 1"), a tombstone value
            /// also has the Data type field set to REG_NONE, the Data size
            /// field set to 0, and the Data offset field set to 0xFFFFFFFF
            const IS_TOMBSTONE = 0x2;
        }
    }

    /// KeySecurity items form a doubly linked list. A key security item may
    /// act as a list header or a list entry (the only difference here is the
    /// meaning of f_link and b_link fields).
    /// When a key security item acts as a list header, flink and blink point to
    /// the first and the last entries of this list respectively. If a list is
    /// empty, flink and blink point to a list header (i.e. to a current cell).
    /// When a key security item acts as a list entry, flink and blink point to
    /// the next and the previous entries of this list respectively. If there is
    /// no next entry in a list, flink points to a list header. If there is no
    /// previous entry in a list, blink points to a list header.
    #[repr(C, packed)]
    #[derive(Debug, Clone, Copy, Zeroable, Pod)]
    pub struct KeySecurity {
        reserved: u16,
        /// offset in bytes relative to the start of hive bins data.
        f_link: Offset,
        /// offset in bytes relative to the start of hive bins data.
        b_link: Offset,
        reference_count: u32,
        security_descriptor_len: u32,
        security_descriptor: ZstPtr,
    }

    /// The [`BigData`] cell type is used to reference data larger than 16344 bytes (when the Minor version field of the base block is greater than 3).
    /// the data `segments` points at is a list of `segments_len` [Offsets](Offset).
    /// A data segment is stored in the data field of a [Cell] pointed by the Data segment offset field. A data segment has the maximum size of 16344 bytes.

    /// `segments` of a [BigData] record, except the last one, always have the maximum size.
    #[repr(C, packed)]
    #[derive(Debug, Clone, Copy, Zeroable, Pod)]
    pub struct BigData {
        segments_len: u16,
        segments: Offset,
    }

    /// Cells contain data which is prefixed with a 2 byte ascii signature
    /// indicating what kind of actual data is stored in this cell
    #[repr(C, packed)]
    #[derive(Debug, Clone, PartialEq, Eq, Copy, Zeroable, Pod)]
    pub struct CellHeader {
        /// 2 byte ASCII signature as represented by constants in CellData
        signature: [u8; 2],
    }

    #[derive(Debug)]
    pub struct Cell<'a> {
        header: CellHeader,
        data: &'a [u8],
    }

    impl<'a> Cell<'a> {
        pub fn into_any_cell(self) -> AnyCell<'a> {
            self.into()
        }
    }

    impl<'a> From<&'a super::RawCell> for Cell<'a> {
        fn from(raw: &'a super::RawCell) -> Self {
            let data = unsafe {
                core::slice::from_raw_parts::<'a, u8>(raw.data.ptr().cast(), raw.size as usize)
            };

            let header = *bytemuck::from_bytes(&data[..2]);
            let data = &data[2..];

            Self { header, data }
        }
    }

    impl<'a> From<Cell<'a>> for AnyCell<'a> {
        fn from(cell: Cell<'a>) -> Self {
            match cell.header {
                CellHeader::FAST_LEAF => {
                    Self::FastLeaf(bytemuck::from_bytes::<Leaf<FastEntry>>(cell.data))
                }
                CellHeader::INDEX_LEAF => {
                    Self::IndexLeaf(bytemuck::from_bytes::<Leaf<Offset>>(cell.data))
                }
                CellHeader::HASH_LEAF => {
                    Self::HashLeaf(bytemuck::from_bytes::<Leaf<HashEntry>>(cell.data))
                }
                CellHeader::INDEX_ROOT => {
                    Self::IndexRoot(bytemuck::from_bytes::<Leaf<Offset>>(cell.data))
                }
                CellHeader::KEY_NODE => Self::KeyNode(bytemuck::from_bytes::<KeyNode>(cell.data)),
                CellHeader::KEY_VALUE => {
                    Self::KeyValue(bytemuck::from_bytes::<KeyValue>(cell.data))
                }
                CellHeader::KEY_SECURITY => {
                    Self::KeySecurity(bytemuck::from_bytes::<KeySecurity>(cell.data))
                }
                CellHeader::BIG_DATA => Self::BigData(bytemuck::from_bytes::<BigData>(cell.data)),
                _ => unimplemented!(),
            }
        }
    }

    /// Helper enum to match against a generic CellData
    pub enum AnyCell<'a> {
        IndexLeaf(&'a Leaf<Offset>),
        FastLeaf(&'a Leaf<FastEntry>),
        HashLeaf(&'a Leaf<HashEntry>),
        IndexRoot(&'a Leaf<Offset>),
        KeyNode(&'a KeyNode),
        KeyValue(&'a KeyValue),
        KeySecurity(&'a KeySecurity),
        BigData(&'a BigData),
    }

    #[repr(C, packed)]
    #[derive(Debug, Clone, Copy, Zeroable, Pod)]
    pub struct KeyNode {
        flags: KeyNodeFlags,
        last_written: u64,
        access: u32,
        parent: Offset,
        subkeys_len: u32,
        volatile_subkeys_len: u32,
        subkeys: Offset,
        volatile_subkeys: Offset,
        key_values_len: u32,
        key_values: Offset,
        key_security: Offset,
        class_name: Offset,
        largest_subkey_len: u32,
        largest_subkey_class_len: u32,
        largest_value_name_len: u32,
        largest_value_data_len: u32,
        work_var: u32,
        key_name_len: u16,
        class_name_len: u16,
        key_name: ZstPtr,
    }

    #[derive(Debug)]
    pub struct SubkeyNameLengthField {
        subkey_name_length: u16,
        virt_flags: VirtualizationControlFlags,
        user_flags: UserFlags,
        debug: DebugFlags,
    }

    impl SubkeyNameLengthField {
        pub fn subkey_name_length(&self) -> u16 {
            self.subkey_name_length
        }

        pub fn virt_flags(&self) -> VirtualizationControlFlags {
            self.virt_flags
        }

        pub fn user_flags(&self) -> UserFlags {
            self.user_flags
        }

        pub fn debug(&self) -> DebugFlags {
            self.debug
        }
    }

    bitflags! {
        #[repr(C)]
        #[derive(Zeroable, Pod)]
        pub struct VirtualizationControlFlags: u8 {
            const DONT_VIRTUALIZE = 0x2;
            const DONT_SILENT_FAIL = 0x4;
            const RECURSE_FLAG = 0x8;

        }
    }

    bitflags! {
        #[repr(C)]
        #[derive(Zeroable, Pod)]
        pub struct KeyNodeFlags: u16 {
            /// Is volatile (not used, a key node on a disk isn't expected to
            /// have this flag set)
            const VOLATILE = 0x0001;
            /// Is the mount point of another hive (a key node on a disk isn't
            /// expected to have this flag set)
                const HIVE_EXIT = 0x0002;
            /// Is the root key for this hive
                const HIVE_ENTRY = 0x0004;
            /// This key can't be deleted
                const NO_DELETE = 0x0008;
            /// This key is a symlink (a target key is specified as a UTF-16LE
            /// string (REG_LINK) in a value named "SymbolicLinkValue", example:
            /// \REGISTRY\MACHINE\SOFTWARE\Classes\Wow6432Node)
                const SYM_LINK = 0x0010;
            /// Key name is an ASCII string, possibly an extended ASCII string
            /// (otherwise it is a UTF-16LE string)
                const COMP_NAME = 0x0020;
            /// Is a predefined handle (a handle is stored in the Number of key
            /// values field)
                const PREDEF_HANDLE = 0x0040;
            /// This key was virtualized at least once
                const VIRTUAL_SOURCE = 0x0080;
            /// Is virtual
                const VIRTUAL_TARGET = 0x0100;
            /// Is a part of a virtual store path
                const VIRTUAL_STORE = 0x0200;
        }
    }

    bitflags! {
        #[repr(C)]
        #[derive(Zeroable, Pod)]
        pub struct DebugFlags: u8 {
            /// This key is opened
            const BREAK_ON_OPEN = 0x01;
            /// This key is deleted
            const BREAK_ON_DELETE = 0x02;
            /// A key security item is changed for this key
            const BREAK_ON_SECURITY_CHANGE = 0x04;
            /// A subkey of this key is created
            const BREAK_ON_CREATE_SUBKEY = 0x08;
            /// A subkey of this key is deleted
            const BREAK_ON_DELETE_SUBKEY = 0x10;
            /// A value is set to this key
            const BREAK_ON_SET_VALUE = 0x20;
            /// A value is deleted from this key
            const BREAK_ON_DELETE_VALUE = 0x40;
            /// This key is virtualized
            const BREAK_ON_KEY_VIRTUALIZE = 0x80;
        }
    }

    bitflags! {
        #[repr(C)]
        #[derive(Zeroable, Pod)]
        pub struct UserFlags: u8 {
            /// Is a 32-bit key: this key was created through the Wow64
            /// subsystem or this key shall not be used by a 64-bit program
            /// (e.g. by a 64-bit driver during the boot)
            const KEY_32_BIT = 0x1;
            /// This key was created by the reflection process (when reflecting
            /// a key from another view)
            const REFLECTION_KEY = 0x2;
            /// Disable registry reflection for this key
            const DISABLE_REFLECTION = 0x4;
            /// In the old location of the User flags field: execute the int 3
            /// instruction on an access to this key (both retail and checked
            /// Windows kernels), this bit was superseded by the Debug field
            /// (see below); in the new location of the User flags field:
            /// disable registry reflection for this key if a corresponding key
            /// exists in another view and it wasn't created by a caller (see
            /// below)
            const TRIGGER_INT3_OR_DISABLE_REFLECTION = 0x8;
        }
    }

    bitflags! {
        #[repr(C)]
        #[derive(Zeroable, Pod)]
        pub struct InnerAccessBits: u8 {
            const PRE_BOOT_REG_INIT = 0x1;
            const POST_BOOT_REG_INIT = 0x2;
        }
    }

    #[repr(C, packed)]
    #[derive(Debug, Clone, Copy, Zeroable, Pod)]
    pub struct AccessBits {
        access_bits: InnerAccessBits,
        layered_key_bit_fields: LayeredKeyBitFields,
    }

    impl From<u32> for AccessBits {
        fn from(value: u32) -> Self {
            let access_bits = InnerAccessBits::from_bits_truncate(value.bit_range(0, 8));
            let layered_key_bit_fields = LayeredKeyBitFields(value.bit_range(8, 16));

            Self {
                access_bits,
                layered_key_bit_fields,
            }
        }
    }

    #[repr(C, packed)]
    #[derive(Debug, Clone, Copy, Zeroable, Pod)]
    struct LayeredKeyBitFields(u8);

    #[repr(C)]
    #[derive(Debug, Clone, Copy, PartialEq, Eq)]
    pub enum LayerSemantics {
        Zero = 0x0,
        IsTombstone,
        IsSupersedeLocal,
        IsSupersedeTree,
    }

    impl LayeredKeyBitFields {
        pub fn inherit_class(&self) -> bool {
            self.0.bit(0)
        }

        pub fn layer_semantics(&self) -> LayerSemantics {
            match BitRange::<u8>::bit_range(&self.0, 6, 8) {
                0 => LayerSemantics::Zero,
                1 => LayerSemantics::IsTombstone,
                2 => LayerSemantics::IsSupersedeLocal,
                3 => LayerSemantics::IsSupersedeTree,
                _ => unreachable!(),
            }
        }
    }

    impl KeyNode {
        pub fn largest_subkey_name_length_vista_xpsp3(&self) -> SubkeyNameLengthField {
            let field = self.largest_subkey_len;
            SubkeyNameLengthField {
                subkey_name_length: field.bit_range(0, 16),
                virt_flags: VirtualizationControlFlags::from_bits_truncate(field.bit_range(16, 20)),
                user_flags: UserFlags::from_bits_truncate(field.bit_range(20, 24)),
                debug: DebugFlags::from_bits_truncate(field.bit_range(24, 32)),
            }
        }
    }

    /// A general view of a *Leaf Cell, which can either be Index, Fast or Hash.
    /// All 3 of these CellData structs share the same common layout and only
    /// differ in what data represents their element
    #[repr(C)]
    #[derive(Debug)]
    pub struct Leaf<T: Sized> {
        len: u16,
        elements: ZstPtr,
        _phantom: PhantomData<T>,
    }

    impl<T: Sized> Clone for Leaf<T> {
        fn clone(&self) -> Self {
            Self {
                len: self.len,
                elements: ZstPtr,
                _phantom: PhantomData,
            }
        }
    }

    impl<T: Sized> Copy for Leaf<T> {}
    unsafe impl<T: Sized> Zeroable for Leaf<T> {}
    unsafe impl<T: Sized + 'static> Pod for Leaf<T> {}

    impl KeyValue {
        pub fn data_type(&self) -> DataType {
            self.data_type.into()
        }

        pub fn small_data(&self) -> Option<[u8; 4]> {
            let data_len = self.data_len;
            let data = self.data.0;

            if data_len.bit(31) && self.data_len() <= 4 {
                let mut ret = [0; 4];
                ret[..self.data_len()]
                    .copy_from_slice(&bytemuck::bytes_of(&data)[..self.data_len()]);

                Some(ret)
            } else {
                None
            }
        }

        pub fn data_len(&self) -> usize {
            let data_len = self.data_len;

            BitRange::<u32>::bit_range(&data_len, 0, 31) as usize
        }
    }

    impl CellHeader {
        const INDEX_LEAF: Self = Self::from_sig(b"li");
        const FAST_LEAF: Self = Self::from_sig(b"lf");
        const HASH_LEAF: Self = Self::from_sig(b"lh");
        const INDEX_ROOT: Self = Self::from_sig(b"ri");
        const KEY_NODE: Self = Self::from_sig(b"nk");
        const KEY_VALUE: Self = Self::from_sig(b"vk");
        const KEY_SECURITY: Self = Self::from_sig(b"sk");
        const BIG_DATA: Self = Self::from_sig(b"bd");

        #[allow(dead_code)]
        const fn from_sig(sig: &[u8; 2]) -> Self {
            Self { signature: *sig }
        }
    }

    #[derive(Debug, Clone, Copy)]
    pub enum DataType {
        None,
        Sz,
        ExpandSz,
        Binary,
        DWordLE,
        DWordBE,
        Link,
        MultiSz,
        ResourceList,
        FullResourceDescriptor,
        ResourceRequirementsList,
        QWordLE,
        Other(u32),
    }

    impl From<u32> for DataType {
        fn from(value: u32) -> Self {
            match value {
                0x00 => Self::None,
                0x01 => Self::Sz,
                0x02 => Self::ExpandSz,
                0x03 => Self::Binary,
                0x04 => Self::DWordLE,
                0x05 => Self::DWordBE,
                0x06 => Self::Link,
                0x07 => Self::MultiSz,
                0x08 => Self::ResourceList,
                0x09 => Self::FullResourceDescriptor,
                0x0a => Self::ResourceRequirementsList,
                0x0b => Self::QWordLE,
                _ => Self::Other(value),
            }
        }
    }

    #[cfg(test)]
    mod test {
        use super::*;
        use core::mem::size_of;

        #[test]
        fn sizes() {
            assert_eq!(size_of::<FastEntry>(), 8, "size_of::<FastEntry>() != 8");
            assert_eq!(size_of::<HashEntry>(), 8, "size_of::<HashEntry>() != 8");
            assert_eq!(size_of::<KeyNode>(), 74, "size_of::<KeyNode>() != 74");
            assert_eq!(size_of::<KeyValue>(), 18, "size_of::<KeyValue>() != 18");
        }
    }
}

use core::borrow::Borrow;

use bytemuck::{Pod, Zeroable};
use cell_data::{AnyCell, Cell, KeyNode};
use error::BaseBlockVerifyError;
use ptr::ZstPtr;

#[cfg(feature = "alloc")]
extern crate alloc;

#[repr(C, packed)]
#[derive(Debug, Copy, Clone, Pod, Zeroable)]
pub struct BaseBlockSequence {
    /// This number is incremented by 1 in the beginning of a write operation on
    /// the primary file
    before: u32,
    /// This number is incremented by 1 at the end of a write operation on the
    /// primary file, a primary sequence number and a secondary sequence number
    /// should be equal after a successful write operation
    after: u32,
}

#[repr(C)]
#[derive(Debug, Copy, Clone, Pod, Zeroable)]
pub struct Offset(u32);

#[derive(Debug)]
pub struct Hive<'a> {
    base_block: &'a BaseBlock,
    hive_bins: &'a [u8],
}

impl<'a> Hive<'a> {
    pub fn new(bytes: &'a [u8]) -> Option<Self> {
        (bytes.len() >= 0x1000)
            .then_some(())
            .and_then(|_| {
                let base_block = bytemuck::from_bytes::<BaseBlock>(bytes);
                let bytes = &bytes[0x1000..];
                let hive_bins_len = base_block.hive_bins_len();

                (bytes.len() >= hive_bins_len).then_some((base_block, &bytes[..hive_bins_len]))
            })
            .map(|(base_block, hive_bins)| Self {
                base_block,
                hive_bins,
            })
    }

    pub fn root_key(&self) -> Option<&KeyNode> {
        let offset = self.base_block.root_cell().offset()?;
        let a = bytemuck::from_bytes::<RawCell>(&self.hive_bins[offset..]);

        let cell = Cell::from(a);

        if let AnyCell::KeyNode(key_node) = cell.into_any_cell() {
            Some(key_node)
        } else {
            None
        }
    }
}

#[repr(C, packed(2))]
#[derive(Debug, Copy, Clone, Pod, Zeroable)]
pub struct BaseBlock {
    /// ASCII string signature equal to `b"regf"`
    signature: u32,
    sequence: BaseBlockSequence,
    /// FILETIME (UTC)
    time_stamp: u64,
    version: Version,
    /// file_type, 0 means primary file
    kind: u32,
    /// 1 means direct memory load
    format: u32,
    /// Offset of a root cell in bytes, relative from the start of the hive bins data
    root_cell: Offset,
    /// Size of the hive bins data in bytes
    hive_bins_len: u32,
    /// Logical sector size of the underlying disk in bytes divided by 512
    clustering_factor: u32,
    /// UTF-16LE string (contains a partial file path to the primary file, or a file name of the primary file), used for debugging purposes
    file_name: [u16; 32],
    win10_fields1: Win10BaseBlock1,
    reserved: [u32; 83],
    /// XOR-32 checksum of the previous 508 bytes
    checksum: u32,
    reserved2: [u32; 0x372],
    win10_fields2: Win10BaseBlock2,
    boot_type: u32,
    boot_recover: u32,
}

#[repr(C, packed)]
#[derive(Debug, Copy, Clone, PartialEq, Eq, Pod, Zeroable)]
pub struct Win10BaseBlock1 {
    rm_id: u128,
    log_id: u128,
    flags: u32,
    tm_id: u128,
    guid_signature: [u8; 4],
    last_reorganized_timestamp: u64,
}

#[repr(C, packed)]
#[derive(Debug, Copy, Clone, PartialEq, Eq, Pod, Zeroable)]
pub struct Win10BaseBlock2 {
    thaw_tm_id: u128,
    thaw_rm_id: u128,
    thaw_log_id: u128,
}

#[repr(C, packed)]
#[derive(Debug, Copy, Clone, PartialEq, Eq, Pod, Zeroable)]
pub struct Version {
    major: u32,
    minor: u32,
}

impl Version {
    pub const V1_3: Self = Self::new(1, 3);
    pub const V1_4: Self = Self::new(1, 4);
    pub const V1_5: Self = Self::new(1, 5);
    pub const V1_6: Self = Self::new(1, 6);

    pub const fn new(major: u32, minor: u32) -> Self {
        Self { major, minor }
    }

    /// returns true if `&self` is any of `[Self::V1_3, Self::V1_4, Self::V1_5, Self::V1_6]`.
    pub fn is_valid(&self) -> bool {
        [Self::V1_3, Self::V1_4, Self::V1_5, Self::V1_6].contains(self)
    }

    pub fn major(&self) -> u32 {
        self.major
    }

    pub fn minor(&self) -> u32 {
        self.minor
    }
}

#[repr(C, packed)]
#[derive(Debug, Copy, Clone, Pod, Zeroable)]
pub struct BinHeader {
    signature: [u8; 4],
    offset: u32,
    size: u32,
    reserved: u64,
    timestamp: u64,
    _spare: u32,
}

#[repr(C, packed)]
#[derive(Debug, Copy, Clone, Pod, Zeroable)]
pub struct RawCell {
    size: i32,
    data: ZstPtr,
}

impl RawCell {
    pub fn size(&self) -> usize {
        self.size.abs() as usize
    }
}

#[repr(C, packed)]
#[derive(Debug, Copy, Clone, Pod, Zeroable)]
pub struct RawBin {
    header: BinHeader,
    cells: ZstPtr,
}

impl RawBin {
    pub fn len(&self) -> usize {
        self.header.size as usize - core::mem::size_of::<BinHeader>()
    }

    pub fn cells_bytes<'a>(&'a self) -> &'a [u8] {
        unsafe { core::slice::from_raw_parts::<'a, u8>(self.cells.ptr().cast(), self.len()) }
    }

    pub fn iter(&self) -> CellIterator {
        CellIterator {
            bin: self,
            offset: 0,
        }
    }
}

#[derive(Debug)]
pub struct CellIterator<'a> {
    bin: &'a RawBin,
    offset: usize,
}

impl<'a> Iterator for CellIterator<'a> {
    type Item = Cell<'a>;

    fn next(&mut self) -> Option<Self::Item> {
        if self.offset >= self.bin.len() {
            None
        } else {
            let bytes = self.bin.cells_bytes();
            let next = bytemuck::try_from_bytes::<RawCell>(&bytes[self.offset..]).ok()?;

            self.offset += next.size();

            Some(next.into())
        }
    }
}

#[cfg(feature = "alloc")]
#[derive(Debug)]
pub struct Bin<'a> {
    raw: &'a RawBin,
    cells: alloc::vec::Vec<Cell<'a>>,
}

#[cfg(feature = "alloc")]
impl<'a> Bin<'a> {
    pub fn header(&self) -> &BinHeader {
        &self.raw.header
    }

    pub fn cells(&self) -> &[Cell] {
        self.cells.as_ref()
    }
}

#[cfg(feature = "alloc")]
impl<'a> From<&'a RawBin> for Bin<'a> {
    fn from(bin: &'a RawBin) -> Self {
        let cells = bin.iter().collect::<alloc::vec::Vec<_>>();

        Self { raw: bin, cells }
    }
}

impl Offset {
    pub fn offset(&self) -> Option<usize> {
        if self.0 == 0xffffffff {
            None
        } else {
            Some(self.0 as usize)
        }
    }
}

impl BaseBlock {
    const SIGNATURE: u32 = 0x66676572;
    const FILE_TYPE_PRIMARY: u32 = 0;
    const BASE_FORMAT_MEMORY: u32 = 1;

    #[cfg(feature = "alloc")]
    pub fn debug_file_name(&self) -> Result<alloc::string::String, alloc::string::FromUtf16Error> {
        use alloc::string::String;
        let name = &self.file_name;
        let name_len = name.iter().take_while(|&&b| b != 0).count();

        String::from_utf16(&self.file_name[..name_len])
    }

    pub fn hive_bins_len(&self) -> usize {
        self.hive_bins_len as usize
    }

    pub fn calculate_checksum(&self) -> u32 {
        let bytes = bytemuck::bytes_of(self);
        // let data = bytemuck::cast_slice::<_, u32>(&bytes[..508]);
        let data = unsafe { core::slice::from_raw_parts(bytes.as_ptr().cast::<u32>(), 127) };

        let checksum = data.iter().clone().fold(0, |a, &b| a ^ b);

        if checksum == 0xffffffff {
            0xfffffffe
        } else if checksum == 0 {
            1u32
        } else {
            checksum
        }
    }

    /// returns either any error encountered when verifying this base block or the dirtiness
    pub fn verify(&self) -> Result<bool, BaseBlockVerifyError> {
        (self.signature == Self::SIGNATURE)
            .then_some(())
            .ok_or(BaseBlockVerifyError::BadSignature)?;
        (self.version.is_valid())
            .then_some(())
            .ok_or(BaseBlockVerifyError::IncompatibleVersion(self.version))?;
        (self.kind == Self::FILE_TYPE_PRIMARY)
            .then_some(())
            .ok_or(BaseBlockVerifyError::BadFileType)?;
        (self.format == Self::BASE_FORMAT_MEMORY)
            .then_some(())
            .ok_or(BaseBlockVerifyError::BadFormat)?;

        (self.clustering_factor == 1)
            .then_some(())
            .ok_or(BaseBlockVerifyError::BadSignature)?;
        Ok(self.dirty())
    }

    /// returns true if this block is dirty
    ///
    /// A hive is considered to be dirty (i.e. requiring recovery) when a base
    /// block in a primary file contains a wrong checksum, or its primary
    /// sequence number doesn't match its secondary sequence number. If a hive
    /// isn't dirty, but a transaction log file (new format) contains subsequent
    /// log entries, they are ignored.
    pub fn dirty(&self) -> bool {
        self.sequence.dirty() || self.calculate_checksum() != self.checksum
    }

    pub fn root_cell(&self) -> Offset {
        self.root_cell
    }
}

impl BaseBlockSequence {
    pub fn dirty(&self) -> bool {
        self.before != self.after
    }

    pub fn get(&self) -> (u32, u32) {
        (self.before, self.after)
    }
}

#[cfg(test)]
mod aligns_and_size {
    use super::*;
    use core::mem::size_of;

    extern crate std;

    #[test]
    fn size_of_base_block() {
        assert_eq!(size_of::<BaseBlock>(), 0x1000);
    }

    #[test]
    fn size_of_base_block_sequence() {
        assert_eq!(size_of::<BaseBlockSequence>(), 8);
    }

    #[test]
    fn size_of_bin_header() {
        assert_eq!(size_of::<BinHeader>(), 32);
    }

    #[test]
    fn size_of_raw_cell() {
        assert_eq!(size_of::<RawCell>(), 4);
    }

    #[test]
    fn offset_of_cells() {
        assert_eq!(bytemuck::offset_of!(RawBin::zeroed(), RawBin, cells), 32);
    }

    #[test]
    fn offset_of_cell_data() {
        assert_eq!(bytemuck::offset_of!(RawCell::zeroed(), RawCell, data), 4);
    }
}
